@@ -47,7 +47,7 @@ struct grub_ohci_hcca
   grub_uint32_t donehead;
 
   grub_uint8_t reserved[116];
-} __attribute__((packed));
+} GRUB_PACKED;
 
 /* OHCI General Transfer Descriptor */
 struct grub_ohci_td
@@ -64,7 +64,7 @@ struct grub_ohci_td
                                * physical address in CPU endian */
   grub_uint32_t tr_index; /* index of TD in transfer */
   grub_uint8_t pad[8 - sizeof (volatile struct grub_ohci_td *)]; /* padding to 32 bytes */
-} __attribute__((packed));
+} GRUB_PACKED;
 
 /* OHCI Endpoint Descriptor.  */
 struct grub_ohci_ed
@@ -73,7 +73,7 @@ struct grub_ohci_ed
   grub_uint32_t td_tail;
   grub_uint32_t td_head;
   grub_uint32_t next_ed;
-} __attribute__((packed));
+} GRUB_PACKED;
 
 typedef volatile struct grub_ohci_td *grub_ohci_td_t;
 typedef volatile struct grub_ohci_ed *grub_ohci_ed_t;
@@ -265,16 +265,20 @@ grub_ohci_pci_iter (grub_pci_device_t dev, grub_pci_id_t pciid,
       addr = grub_pci_make_address (dev, GRUB_PCI_REG_ADDRESS_REG0);
       base = grub_pci_read (addr);
 
-#if 0
-      /* Stop if there is no IO space base address defined.  */
-      if (! (base & 1))
-	return 0;
-#endif
+      base &= GRUB_PCI_ADDR_MEM_MASK;
+      if (!base)
+	{
+	  grub_dprintf ("ehci",
+			"EHCI: EHCI is not mapper\n");
+	  return 0;
+	}
 
       /* Set bus master - needed for coreboot, VMware, broken BIOSes etc. */
       addr = grub_pci_make_address (dev, GRUB_PCI_REG_COMMAND);
       grub_pci_write_word(addr,
-          GRUB_PCI_COMMAND_BUS_MASTER | grub_pci_read_word(addr));
+			  GRUB_PCI_COMMAND_MEM_ENABLED
+			  | GRUB_PCI_COMMAND_BUS_MASTER
+			  | grub_pci_read_word(addr));
 
       grub_dprintf ("ohci", "class=0x%02x 0x%02x interface 0x%02x\n",
 		    class, subclass, interf);
@@ -310,7 +314,7 @@ grub_ohci_pci_iter (grub_pci_device_t dev, grub_pci_id_t pciid,
   grub_memset ((void *) o->ed_ctrl, 0, sizeof (struct grub_ohci_ed)
 	       * GRUB_OHCI_CTRL_EDS);
   for (j=0; j < GRUB_OHCI_CTRL_EDS; j++)
-    o->ed_ctrl[j].target = grub_cpu_to_le32 (1 << 14); /* skip */
+    o->ed_ctrl[j].target = grub_cpu_to_le32_compile_time (1 << 14); /* skip */
     
   grub_dprintf ("ohci", "EDs-C: chunk=%p, virt=%p, phys=0x%02x\n",
                 o->ed_ctrl_chunk, o->ed_ctrl, o->ed_ctrl_addr);
@@ -325,7 +329,7 @@ grub_ohci_pci_iter (grub_pci_device_t dev, grub_pci_id_t pciid,
   /* Preset EDs */
   grub_memset ((void*)o->ed_bulk, 0, sizeof(struct grub_ohci_ed) * GRUB_OHCI_BULK_EDS);
   for (j=0; j < GRUB_OHCI_BULK_EDS; j++)
-    o->ed_bulk[j].target = grub_cpu_to_le32 (1 << 14); /* skip */
+    o->ed_bulk[j].target = grub_cpu_to_le32_compile_time (1 << 14); /* skip */
 
   grub_dprintf ("ohci", "EDs-B: chunk=%p, virt=%p, phys=0x%02x\n",
                 o->ed_bulk_chunk, o->ed_bulk, o->ed_bulk_addr);
@@ -746,7 +750,10 @@ grub_ohci_setup_transfer (grub_usb_controller_t dev,
     {
       cdata->td_head_virt = grub_ohci_alloc_td (o);
       if (!cdata->td_head_virt)
-        return GRUB_USB_ERR_INTERNAL; /* We don't need de-allocate ED */
+	{
+	  grub_free (cdata);
+	  return GRUB_USB_ERR_INTERNAL; /* We don't need de-allocate ED */
+	}
       /* We can set td_head only when ED is not active, i.e.
        * when it is newly allocated. */
       cdata->ed_virt->td_head
@@ -1216,7 +1223,7 @@ grub_ohci_cancel_transfer (grub_usb_controller_t dev,
   return GRUB_USB_ERR_NONE;
 }
 
-static grub_err_t
+static grub_usb_err_t
 grub_ohci_portstatus (grub_usb_controller_t dev,
 		      unsigned int port, unsigned int enable)
 {
@@ -1236,11 +1243,11 @@ grub_ohci_portstatus (grub_usb_controller_t dev,
        while ((grub_ohci_readreg32 (o, GRUB_OHCI_REG_RHUBPORT + port)
                & (1 << 1)))
          if (grub_get_time_ms () > endtime)
-           return grub_error (GRUB_ERR_IO, "OHCI Timed out - disable");
+           return GRUB_USB_ERR_TIMEOUT;
 
        grub_dprintf ("ohci", "end of portstatus=0x%02x\n",
          grub_ohci_readreg32 (o, GRUB_OHCI_REG_RHUBPORT + port));
-       return GRUB_ERR_NONE;
+       return GRUB_USB_ERR_NONE;
      }
      
    /* OHCI does one reset signal 10ms long but USB spec.
@@ -1257,7 +1264,7 @@ grub_ohci_portstatus (grub_usb_controller_t dev,
        while (! (grub_ohci_readreg32 (o, GRUB_OHCI_REG_RHUBPORT + port)
                & GRUB_OHCI_SET_PORT_RESET_STATUS_CHANGE))
          if (grub_get_time_ms () > endtime)
-           return grub_error (GRUB_ERR_IO, "OHCI Timed out - reset");
+           return GRUB_USB_ERR_TIMEOUT;
 
        /* End the reset signaling - reset the reset status change */
        grub_ohci_writereg32 (o, GRUB_OHCI_REG_RHUBPORT + port,
@@ -1275,7 +1282,7 @@ grub_ohci_portstatus (grub_usb_controller_t dev,
    while (! (grub_ohci_readreg32 (o, GRUB_OHCI_REG_RHUBPORT + port)
            & (1 << 1)))
      if (grub_get_time_ms () > endtime)
-       return grub_error (GRUB_ERR_IO, "OHCI Timed out - enable");
+       return GRUB_USB_ERR_TIMEOUT;
 
    /* Reset bit Connect Status Change */
    grub_ohci_writereg32 (o, GRUB_OHCI_REG_RHUBPORT + port,
@@ -1287,7 +1294,7 @@ grub_ohci_portstatus (grub_usb_controller_t dev,
    grub_dprintf ("ohci", "end of portstatus=0x%02x\n",
 		 grub_ohci_readreg32 (o, GRUB_OHCI_REG_RHUBPORT + port));
  
-   return GRUB_ERR_NONE;
+   return GRUB_USB_ERR_NONE;
 }
 
 static grub_usb_speed_t
@@ -1345,10 +1352,10 @@ grub_ohci_fini_hw (int noreturn __attribute__ ((unused)))
       /* Set skip in all EDs */
       if (o->ed_bulk)
         for (i=0; i < GRUB_OHCI_BULK_EDS; i++)
-          o->ed_bulk[i].target |= grub_cpu_to_le32 (1 << 14); /* skip */
+          o->ed_bulk[i].target |= grub_cpu_to_le32_compile_time (1 << 14); /* skip */
       if (o->ed_ctrl)
         for (i=0; i < GRUB_OHCI_CTRL_EDS; i++)
-          o->ed_ctrl[i].target |= grub_cpu_to_le32 (1 << 14); /* skip */
+          o->ed_ctrl[i].target |= grub_cpu_to_le32_compile_time (1 << 14); /* skip */
 
       /* We should wait for next SOF to be sure that all EDs are
        * unaccessed by OHCI. But OHCI can be non-functional, so
