@@ -58,7 +58,7 @@ GRUB_MOD_LICENSE ("GPLv3+");
 	(grub_le_to_cpu32 (data->sblock.log2_block_size) + 10)
 
 /* The size of an ext2 block in bytes.  */
-#define EXT2_BLOCK_SIZE(data)		(1 << LOG2_BLOCK_SIZE (data))
+#define EXT2_BLOCK_SIZE(data)		(1U << LOG2_BLOCK_SIZE (data))
 
 /* The revision level.  */
 #define EXT2_REVISION(data)	grub_le_to_cpu32 (data->sblock.revision_level)
@@ -100,6 +100,7 @@ GRUB_MOD_LICENSE ("GPLv3+");
 #define EXT2_FEATURE_INCOMPAT_META_BG		0x0010
 #define EXT4_FEATURE_INCOMPAT_EXTENTS		0x0040 /* Extents used */
 #define EXT4_FEATURE_INCOMPAT_64BIT		0x0080
+#define EXT4_FEATURE_INCOMPAT_MMP		0x0100
 #define EXT4_FEATURE_INCOMPAT_FLEX_BG		0x0200
 
 /* The set of back-incompatible features this driver DOES support. Add (OR)
@@ -107,6 +108,7 @@ GRUB_MOD_LICENSE ("GPLv3+");
 #define EXT2_DRIVER_SUPPORTED_INCOMPAT ( EXT2_FEATURE_INCOMPAT_FILETYPE \
                                        | EXT4_FEATURE_INCOMPAT_EXTENTS  \
                                        | EXT4_FEATURE_INCOMPAT_FLEX_BG \
+                                       | EXT2_FEATURE_INCOMPAT_META_BG \
                                        | EXT4_FEATURE_INCOMPAT_64BIT)
 /* List of rationales for the ignored "incompatible" features:
  * needs_recovery: Not really back-incompatible - was added as such to forbid
@@ -114,8 +116,13 @@ GRUB_MOD_LICENSE ("GPLv3+");
  *                 journal because they will ignore the journal, but the next
  *                 ext3 driver to mount the volume will find the journal and
  *                 replay it, potentially corrupting the metadata written by
- *                 the ext2 drivers. Safe to ignore for this RO driver.  */
-#define EXT2_DRIVER_IGNORED_INCOMPAT ( EXT3_FEATURE_INCOMPAT_RECOVER )
+ *                 the ext2 drivers. Safe to ignore for this RO driver.
+ * mmp:            Not really back-incompatible - was added as such to
+ *                 avoid multiple read-write mounts. Safe to ignore for this
+ *                 RO driver.
+ */
+#define EXT2_DRIVER_IGNORED_INCOMPAT ( EXT3_FEATURE_INCOMPAT_RECOVER \
+				     | EXT4_FEATURE_INCOMPAT_MMP)
 
 
 #define EXT3_JOURNAL_MAGIC_NUMBER	0xc03b3998U
@@ -247,6 +254,7 @@ struct ext2_dirent
 {
   grub_uint32_t inode;
   grub_uint16_t direntlen;
+#define MAX_NAMELEN 255
   grub_uint8_t namelen;
   grub_uint8_t filetype;
 };
@@ -330,25 +338,78 @@ static grub_dl_t my_mod;
 
 
 
+/* Check is a = b^x for some x.  */
+static inline int
+is_power_of (grub_uint64_t a, grub_uint32_t b)
+{
+  grub_uint64_t c;
+  /* Prevent overflow assuming b < 8.  */
+  if (a >= (1LL << 60))
+    return 0;
+  for (c = 1; c <= a; c *= b);
+  return (c == a);
+}
+
+
+static inline int
+group_has_super_block (struct grub_ext2_data *data, grub_uint64_t group)
+{
+  if (!(data->sblock.feature_ro_compat
+	& grub_cpu_to_le32_compile_time(EXT2_FEATURE_RO_COMPAT_SPARSE_SUPER)))
+    return 1;
+  /* Algorithm looked up in Linux source.  */
+  if (group <= 1)
+    return 1;
+  /* Even number is never a power of odd number.  */
+  if (!(group & 1))
+    return 0;
+  return (is_power_of(group, 7) || is_power_of(group, 5) ||
+	  is_power_of(group, 3));
+}
+
 /* Read into BLKGRP the blockgroup descriptor of blockgroup GROUP of
    the mounted filesystem DATA.  */
 inline static grub_err_t
-grub_ext2_blockgroup (struct grub_ext2_data *data, int group,
+grub_ext2_blockgroup (struct grub_ext2_data *data, grub_uint64_t group,
 		      struct grub_ext2_block_group *blkgrp)
 {
+  grub_uint64_t full_offset = (group << data->log_group_desc_size);
+  grub_uint64_t block, offset;
+  block = (full_offset >> LOG2_BLOCK_SIZE (data));
+  offset = (full_offset & ((1 << LOG2_BLOCK_SIZE (data)) - 1));
+  if ((data->sblock.feature_incompat
+       & grub_cpu_to_le32_compile_time (EXT2_FEATURE_INCOMPAT_META_BG))
+      && block >= grub_le_to_cpu32(data->sblock.first_meta_bg))
+    {
+      grub_uint64_t first_block_group;
+      /* Find the first block group for which a descriptor
+	 is stored in given block. */
+      first_block_group = (block << (LOG2_BLOCK_SIZE (data)
+				     - data->log_group_desc_size));
+
+      block = (first_block_group
+	       * grub_le_to_cpu32(data->sblock.blocks_per_group));
+
+      if (group_has_super_block (data, first_block_group))
+	block++;
+    }
+  else
+    /* Superblock.  */
+    block++;
   return grub_disk_read (data->disk,
-                         ((grub_le_to_cpu32 (data->sblock.first_data_block) + 1)
-                          << LOG2_EXT2_BLOCK_SIZE (data)),
-			 group << data->log_group_desc_size,
+                         ((grub_le_to_cpu32 (data->sblock.first_data_block)
+			   + block)
+                          << LOG2_EXT2_BLOCK_SIZE (data)), offset,
 			 sizeof (struct grub_ext2_block_group), blkgrp);
 }
 
 static struct grub_ext4_extent_header *
-grub_ext4_find_leaf (struct grub_ext2_data *data, grub_properly_aligned_t *buf,
+grub_ext4_find_leaf (struct grub_ext2_data *data,
                      struct grub_ext4_extent_header *ext_block,
                      grub_uint32_t fileblock)
 {
   struct grub_ext4_extent_idx *index;
+  void *buf = NULL;
 
   while (1)
     {
@@ -357,8 +418,8 @@ grub_ext4_find_leaf (struct grub_ext2_data *data, grub_properly_aligned_t *buf,
 
       index = (struct grub_ext4_extent_idx *) (ext_block + 1);
 
-      if (grub_le_to_cpu16(ext_block->magic) != EXT4_EXT_MAGIC)
-        return 0;
+      if (ext_block->magic != grub_cpu_to_le16_compile_time (EXT4_EXT_MAGIC))
+	goto fail;
 
       if (ext_block->depth == 0)
         return ext_block;
@@ -370,17 +431,24 @@ grub_ext4_find_leaf (struct grub_ext2_data *data, grub_properly_aligned_t *buf,
         }
 
       if (--i < 0)
-        return 0;
+	goto fail;
 
       block = grub_le_to_cpu16 (index[i].leaf_hi);
       block = (block << 32) | grub_le_to_cpu32 (index[i].leaf);
+      if (!buf)
+	buf = grub_malloc (EXT2_BLOCK_SIZE(data));
+      if (!buf)
+	goto fail;
       if (grub_disk_read (data->disk,
                           block << LOG2_EXT2_BLOCK_SIZE (data),
                           0, EXT2_BLOCK_SIZE(data), buf))
-        return 0;
+	goto fail;
 
-      ext_block = (struct grub_ext4_extent_header *) buf;
+      ext_block = buf;
     }
+ fail:
+  grub_free (buf);
+  return 0;
 }
 
 static grub_disk_addr_t
@@ -388,20 +456,21 @@ grub_ext2_read_block (grub_fshelp_node_t node, grub_disk_addr_t fileblock)
 {
   struct grub_ext2_data *data = node->data;
   struct grub_ext2_inode *inode = &node->inode;
-  grub_disk_addr_t blknr = -1;
   unsigned int blksz = EXT2_BLOCK_SIZE (data);
+  grub_disk_addr_t blksz_quarter = blksz / 4;
   int log2_blksz = LOG2_EXT2_BLOCK_SIZE (data);
+  int log_perblock = log2_blksz + 9 - 2;
+  grub_uint32_t indir;
+  int shift;
 
   if (inode->flags & grub_cpu_to_le32_compile_time (EXT4_EXTENTS_FLAG))
     {
-      GRUB_PROPERLY_ALIGNED_ARRAY (buf, EXT2_BLOCK_SIZE(data));
       struct grub_ext4_extent_header *leaf;
       struct grub_ext4_extent *ext;
       int i;
+      grub_disk_addr_t ret;
 
-      leaf = grub_ext4_find_leaf (data, buf,
-                                  (struct grub_ext4_extent_header *) inode->blocks.dir_blocks,
-                                  fileblock);
+      leaf = grub_ext4_find_leaf (data, (struct grub_ext4_extent_header *) inode->blocks.dir_blocks, fileblock);
       if (! leaf)
         {
           grub_error (GRUB_ERR_BAD_FS, "invalid extent");
@@ -419,7 +488,7 @@ grub_ext2_read_block (grub_fshelp_node_t node, grub_disk_addr_t fileblock)
         {
           fileblock -= grub_le_to_cpu32 (ext[i].block);
           if (fileblock >= grub_le_to_cpu16 (ext[i].len))
-            return 0;
+	    ret = 0;
           else
             {
               grub_disk_addr_t start;
@@ -427,98 +496,69 @@ grub_ext2_read_block (grub_fshelp_node_t node, grub_disk_addr_t fileblock)
               start = grub_le_to_cpu16 (ext[i].start_hi);
               start = (start << 32) + grub_le_to_cpu32 (ext[i].start);
 
-              return fileblock + start;
+              ret = fileblock + start;
             }
         }
       else
         {
           grub_error (GRUB_ERR_BAD_FS, "something wrong with extent");
-          return -1;
+	  ret = -1;
         }
+
+      if (leaf != (struct grub_ext4_extent_header *) inode->blocks.dir_blocks)
+	grub_free (leaf);
+
+      return ret;
     }
+
   /* Direct blocks.  */
   if (fileblock < INDIRECT_BLOCKS)
-    blknr = grub_le_to_cpu32 (inode->blocks.dir_blocks[fileblock]);
+    return grub_le_to_cpu32 (inode->blocks.dir_blocks[fileblock]);
+  fileblock -= INDIRECT_BLOCKS;
   /* Indirect.  */
-  else if (fileblock < INDIRECT_BLOCKS + blksz / 4)
+  if (fileblock < blksz_quarter)
     {
-      grub_uint32_t indir[blksz / 4];
-
-      if (grub_disk_read (data->disk,
-			  ((grub_disk_addr_t)
-			   grub_le_to_cpu32 (inode->blocks.indir_block))
-			  << log2_blksz,
-			  0, blksz, indir))
-	return grub_errno;
-
-      blknr = grub_le_to_cpu32 (indir[fileblock - INDIRECT_BLOCKS]);
+      indir = inode->blocks.indir_block;
+      shift = 0;
+      goto indirect;
     }
+  fileblock -= blksz_quarter;
   /* Double indirect.  */
-  else if (fileblock < INDIRECT_BLOCKS
-	   + blksz / 4 * ((grub_disk_addr_t) blksz / 4 + 1))
+  if (fileblock < blksz_quarter * blksz_quarter)
     {
-      int log_perblock = log2_blksz + 9 - 2;
-      grub_disk_addr_t rblock = fileblock - (INDIRECT_BLOCKS
-					     + blksz / 4);
-      grub_uint32_t indir[blksz / 4];
-
-      if (grub_disk_read (data->disk,
-			  ((grub_disk_addr_t)
-			   grub_le_to_cpu32 (inode->blocks.double_indir_block))
-			  << log2_blksz,
-			  0, blksz, indir))
-	return grub_errno;
-
-      if (grub_disk_read (data->disk,
-			  ((grub_disk_addr_t)
-			   grub_le_to_cpu32 (indir[rblock >> log_perblock]))
-			  << log2_blksz,
-			  0, blksz, indir))
-	return grub_errno;
-
-
-      blknr = grub_le_to_cpu32 (indir[rblock & ((1 << log_perblock) - 1)]);
+      indir = inode->blocks.double_indir_block;
+      shift = 1;
+      goto indirect;
     }
-  /* triple indirect.  */
-  else if (fileblock < INDIRECT_BLOCKS + blksz / 4 * ((grub_disk_addr_t) blksz / 4 + 1)
-	   + ((grub_disk_addr_t) blksz / 4) * ((grub_disk_addr_t) blksz / 4)
-	   * ((grub_disk_addr_t) blksz / 4 + 1))
+  fileblock -= blksz_quarter * blksz_quarter;
+  /* Triple indirect.  */
+  if (fileblock < blksz_quarter * blksz_quarter * (blksz_quarter + 1))
     {
-      int log_perblock = log2_blksz + 9 - 2;
-      grub_disk_addr_t rblock = fileblock - (INDIRECT_BLOCKS + blksz / 4
-					     * (blksz / 4 + 1));
-      grub_uint32_t indir[blksz / 4];
-
-      if (grub_disk_read (data->disk,
-			  ((grub_disk_addr_t)
-			   grub_le_to_cpu32 (inode->blocks.triple_indir_block))
-			  << log2_blksz,
-			  0, blksz, indir))
-	return grub_errno;
-
-      if (grub_disk_read (data->disk,
-			  ((grub_disk_addr_t)
-			   grub_le_to_cpu32 (indir[(rblock >> log_perblock) >> log_perblock]))
-			  << log2_blksz,
-			  0, blksz, indir))
-	return grub_errno;
-
-      if (grub_disk_read (data->disk,
-			  ((grub_disk_addr_t)
-			   grub_le_to_cpu32 (indir[(rblock >> log_perblock) & ((1 << log_perblock) - 1)]))
-			  << log2_blksz,
-			  0, blksz, indir))
-	return grub_errno;
-
-      blknr = grub_le_to_cpu32 (indir[rblock  & ((1 << log_perblock) - 1)]);
-    }
-  else
-    {
-      grub_error (GRUB_ERR_NOT_IMPLEMENTED_YET,
-		  "ext2fs doesn't support quadruple indirect blocks");
+      indir = inode->blocks.triple_indir_block;
+      shift = 2;
+      goto indirect;
     }
+  grub_error (GRUB_ERR_BAD_FS,
+	      "ext2fs doesn't support quadruple indirect blocks");
+  return -1;
 
-  return blknr;
+indirect:
+  do {
+    /* If the indirect block is zero, all child blocks are absent
+       (i.e. filled with zeros.) */
+    if (indir == 0)
+      return 0;
+    if (grub_disk_read (data->disk,
+			((grub_disk_addr_t) grub_le_to_cpu32 (indir))
+			<< log2_blksz,
+			((fileblock >> (log_perblock * shift))
+			 & ((1 << log_perblock) - 1))
+			* sizeof (indir),
+			sizeof (indir), &indir))
+      return -1;
+  } while (shift--);
+
+  return grub_le_to_cpu32 (indir);
 }
 
 /* Read LEN bytes from the file described by DATA starting with byte
@@ -597,7 +637,12 @@ grub_ext2_mount (grub_disk_t disk)
 
   /* Make sure this is an ext2 filesystem.  */
   if (data->sblock.magic != grub_cpu_to_le16_compile_time (EXT2_MAGIC)
-      || grub_le_to_cpu32 (data->sblock.log2_block_size) >= 16)
+      || grub_le_to_cpu32 (data->sblock.log2_block_size) >= 16
+      || data->sblock.inodes_per_group == 0
+      /* 20 already means 1GiB blocks. We don't want to deal with blocks overflowing int32. */
+      || grub_le_to_cpu32 (data->sblock.log2_block_size) > 20
+      || EXT2_INODE_SIZE (data) == 0
+      || EXT2_BLOCK_SIZE (data) / EXT2_INODE_SIZE (data) == 0)
     {
       grub_error (GRUB_ERR_BAD_FS, "not an ext2 filesystem");
       goto fail;
@@ -670,10 +715,10 @@ grub_ext2_read_symlink (grub_fshelp_node_t node)
   /* If the filesize of the symlink is bigger than
      60 the symlink is stored in a separate block,
      otherwise it is stored in the inode.  */
-  if (grub_le_to_cpu32 (diro->inode.size) <= 60)
-    grub_strncpy (symlink,
-		  diro->inode.symlink,
-		  grub_le_to_cpu32 (diro->inode.size));
+  if (grub_le_to_cpu32 (diro->inode.size) <= sizeof (diro->inode.symlink))
+    grub_memcpy (symlink,
+		 diro->inode.symlink,
+		 grub_le_to_cpu32 (diro->inode.size));
   else
     {
       grub_ext2_read_file (diro, 0, 0, 0,
@@ -719,7 +764,7 @@ grub_ext2_iterate_dir (grub_fshelp_node_t dir,
 
       if (dirent.inode != 0 && dirent.namelen != 0)
 	{
-	  char filename[dirent.namelen + 1];
+	  char filename[MAX_NAMELEN + 1];
 	  struct grub_fshelp_node *fdiro;
 	  enum grub_fshelp_filetype type = GRUB_FSHELP_UNKNOWN;
 
