@@ -54,7 +54,7 @@ struct grub_iso9660_voldesc
   grub_uint8_t type;
   grub_uint8_t magic[5];
   grub_uint8_t version;
-} __attribute__ ((packed));
+} GRUB_PACKED;
 
 struct grub_iso9660_date2
 {
@@ -65,7 +65,7 @@ struct grub_iso9660_date2
   grub_uint8_t minute;
   grub_uint8_t second;
   grub_uint8_t offset;
-} __attribute__ ((packed));
+} GRUB_PACKED;
 
 /* A directory entry.  */
 struct grub_iso9660_dir
@@ -79,8 +79,9 @@ struct grub_iso9660_dir
   struct grub_iso9660_date2 mtime;
   grub_uint8_t flags;
   grub_uint8_t unused2[6];
+#define MAX_NAMELEN 255
   grub_uint8_t namelen;
-} __attribute__ ((packed));
+} GRUB_PACKED;
 
 struct grub_iso9660_date
 {
@@ -92,7 +93,7 @@ struct grub_iso9660_date
   grub_uint8_t second[2];
   grub_uint8_t hundredth[2];
   grub_uint8_t offset;
-} __attribute__ ((packed));
+} GRUB_PACKED;
 
 /* The primary volume descriptor.  Only little endian is used.  */
 struct grub_iso9660_primary_voldesc
@@ -111,7 +112,7 @@ struct grub_iso9660_primary_voldesc
   grub_uint8_t unused6[624];
   struct grub_iso9660_date created;
   struct grub_iso9660_date modified;
-} __attribute__ ((packed));
+} GRUB_PACKED;
 
 /* A single entry in the path table.  */
 struct grub_iso9660_path
@@ -121,7 +122,7 @@ struct grub_iso9660_path
   grub_uint32_t first_sector;
   grub_uint16_t parentdir;
   grub_uint8_t name[0];
-} __attribute__ ((packed));
+} GRUB_PACKED;
 
 /* An entry in the System Usage area of the directory entry.  */
 struct grub_iso9660_susp_entry
@@ -130,7 +131,7 @@ struct grub_iso9660_susp_entry
   grub_uint8_t len;
   grub_uint8_t version;
   grub_uint8_t data[0];
-} __attribute__ ((packed));
+} GRUB_PACKED;
 
 /* The CE entry.  This is used to describe the next block where data
    can be found.  */
@@ -143,7 +144,7 @@ struct grub_iso9660_susp_ce
   grub_uint32_t off_be;
   grub_uint32_t len;
   grub_uint32_t len_be;
-} __attribute__ ((packed));
+} GRUB_PACKED;
 
 struct grub_iso9660_data
 {
@@ -328,11 +329,11 @@ grub_iso9660_convert_string (grub_uint8_t *us, int len)
 {
   char *p;
   int i;
-  grub_uint16_t t[len];
+  grub_uint16_t t[MAX_NAMELEN / 2 + 1];
 
   p = grub_malloc (len * GRUB_MAX_UTF8_PER_UTF16 + 1);
   if (! p)
-    return p;
+    return NULL;
 
   for (i=0; i<len; i++)
     t[i] = grub_be_to_cpu16 (grub_get_unaligned16 (us + 2 * i));
@@ -535,8 +536,8 @@ add_part (struct iterate_dir_ctx *ctx,
   if (! ctx->symlink)
     return;
 
-  ctx->symlink[size] = 0;
-  grub_strncat (ctx->symlink, part, len2);
+  grub_memcpy (ctx->symlink + size, part, len2);
+  ctx->symlink[size + len2] = 0;  
 }
 
 static grub_err_t
@@ -557,20 +558,19 @@ susp_iterate_dir (struct grub_iso9660_susp_entry *entry,
 	ctx->filename = (char *) "..";
       else if (entry->len >= 5)
 	{
-	  grub_size_t size = 1, csize = 1;
+	  grub_size_t off = 0, csize = 1;
 	  char *old;
-	  csize = size = entry->len - 5;
+	  csize = entry->len - 5;
 	  old = ctx->filename;
 	  if (ctx->filename_alloc)
 	    {
-	      size += grub_strlen (ctx->filename);
-	      ctx->filename = grub_realloc (ctx->filename, size + 1);
+	      off = grub_strlen (ctx->filename);
+	      ctx->filename = grub_realloc (ctx->filename, csize + off + 1);
 	    }
 	  else
 	    {
-	      ctx->filename_alloc = 1;
-	      ctx->filename = grub_zalloc (size + 1);
-	      ctx->filename[0] = 0;
+	      off = 0;
+	      ctx->filename = grub_zalloc (csize + 1);
 	    }
 	  if (!ctx->filename)
 	    {
@@ -578,8 +578,8 @@ susp_iterate_dir (struct grub_iso9660_susp_entry *entry,
 	      return grub_errno;
 	    }
 	  ctx->filename_alloc = 1;
-	  grub_strncat (ctx->filename, (char *) &entry->data[1], csize);
-	  ctx->filename[size] = '\0';
+	  grub_memcpy (ctx->filename + off, (char *) &entry->data[1], csize);
+	  ctx->filename[off + csize] = '\0';
 	}
     }
   /* The mode information (st_mode).  */
@@ -680,7 +680,7 @@ grub_iso9660_iterate_dir (grub_fshelp_node_t dir,
 	}
 
       {
-	char name[dirent.namelen + 1];
+	char name[MAX_NAMELEN + 1];
 	int nameoffset = offset + sizeof (dirent);
 	struct grub_fshelp_node *node;
 	int sua_off = (sizeof (dirent) + dirent.namelen + 1
@@ -750,19 +750,15 @@ grub_iso9660_iterate_dir (grub_fshelp_node_t dir,
 
         if (dir->data->joliet && !ctx.filename)
           {
-            char *oldname, *semicolon;
+            char *semicolon;
 
-            oldname = name;
             ctx.filename = grub_iso9660_convert_string
-                  ((grub_uint8_t *) oldname, dirent.namelen >> 1);
+                  ((grub_uint8_t *) name, dirent.namelen >> 1);
 
 	    semicolon = grub_strrchr (ctx.filename, ';');
 	    if (semicolon)
 	      *semicolon = '\0';
 
-            if (ctx.filename_alloc)
-              grub_free (oldname);
-
             ctx.filename_alloc = 1;
           }
 
@@ -959,14 +955,15 @@ grub_iso9660_read (grub_file_t file, char *buf, grub_size_t len)
 {
   struct grub_iso9660_data *data =
     (struct grub_iso9660_data *) file->data;
+  grub_err_t err;
 
   /* XXX: The file is stored in as a single extent.  */
   data->disk->read_hook = file->read_hook;
   data->disk->read_hook_data = file->read_hook_data;
-  read_node (data->node, file->offset, len, buf);
+  err = read_node (data->node, file->offset, len, buf);
   data->disk->read_hook = NULL;
 
-  if (grub_errno)
+  if (err || grub_errno)
     return -1;
 
   return len;
