@@ -69,7 +69,6 @@ static grub_addr_t prot_mode_target;
 static void *initrd_mem;
 static grub_addr_t initrd_mem_target;
 static grub_size_t prot_init_space;
-static grub_uint32_t initrd_pages;
 static struct grub_relocator *relocator = NULL;
 static void *efi_mmap_buf;
 static grub_size_t maximal_cmdline_size;
@@ -87,7 +86,7 @@ struct idt_descriptor
 {
   grub_uint16_t limit;
   void *base;
-} __attribute__ ((packed));
+} GRUB_PACKED;
 
 static struct idt_descriptor idt_desc =
   {
@@ -198,6 +197,9 @@ allocate_pages (grub_size_t prot_size, grub_size_t *align,
 {
   grub_err_t err;
 
+  if (prot_size == 0)
+    prot_size = 1;
+
   prot_size = page_align (prot_size);
 
   /* Initialize the memory pointers with NULL for convenience.  */
@@ -246,8 +248,8 @@ allocate_pages (grub_size_t prot_size, grub_size_t *align,
     prot_mode_target = get_physical_target_address (ch);
   }
 
-  grub_dprintf ("linux", "prot_mode_mem = %lx, prot_mode_target = %lx, prot_size = %x\n",
-                (unsigned long) prot_mode_mem, (unsigned long) prot_mode_target,
+  grub_dprintf ("linux", "prot_mode_mem = %p, prot_mode_target = %lx, prot_size = %x\n",
+                prot_mode_mem, (unsigned long) prot_mode_target,
 		(unsigned) prot_size);
   return GRUB_ERR_NONE;
 
@@ -338,9 +340,11 @@ grub_linux_setup_video (struct linux_kernel_params *params)
 	case GRUB_VIDEO_DRIVER_CIRRUS:
 	case GRUB_VIDEO_DRIVER_BOCHS:
 	case GRUB_VIDEO_DRIVER_RADEON_FULOONG2E:
+	case GRUB_VIDEO_DRIVER_RADEON_YEELOONG3A:
 	case GRUB_VIDEO_DRIVER_IEEE1275:
 	case GRUB_VIDEO_DRIVER_COREBOOT:
 	  /* Make gcc happy. */
+	case GRUB_VIDEO_DRIVER_XEN:
 	case GRUB_VIDEO_DRIVER_SDL:
 	case GRUB_VIDEO_DRIVER_NONE:
 	case GRUB_VIDEO_ADAPTER_CAPTURE:
@@ -421,36 +425,15 @@ grub_linux_boot_mmap_find (grub_uint64_t addr, grub_uint64_t size,
   return 1;
 }
 
+/* GRUB types conveniently match E820 types.  */
 static int
 grub_linux_boot_mmap_fill (grub_uint64_t addr, grub_uint64_t size,
 			   grub_memory_type_t type, void *data)
 {
   struct grub_linux_boot_ctx *ctx = data;
 
-  grub_uint32_t e820_type;
-  switch (type)
-    {
-    case GRUB_MEMORY_AVAILABLE:
-      e820_type = GRUB_E820_RAM;
-      break;
-
-    case GRUB_MEMORY_ACPI:
-      e820_type = GRUB_E820_ACPI;
-      break;
-
-    case GRUB_MEMORY_NVS:
-      e820_type = GRUB_E820_NVS;
-      break;
-
-    case GRUB_MEMORY_BADRAM:
-      e820_type = GRUB_E820_BADRAM;
-      break;
-
-    default:
-      e820_type = GRUB_E820_RESERVED;
-    }
   if (grub_e820_add_region (ctx->params->e820_map, &ctx->e820_num,
-			    addr, size, e820_type))
+			    addr, size, type))
     return 1;
 
   return 0;
@@ -551,9 +534,9 @@ grub_linux_boot (void)
 	    || grub_strcmp (term->name, "console") == 0
 	    || grub_strcmp (term->name, "ofconsole") == 0)
 	  {
-	    grub_uint16_t pos = grub_term_getxy (term);
-	    linux_params.video_cursor_x = pos >> 8;
-	    linux_params.video_cursor_y = pos & 0xff;
+	    struct grub_term_coordinate pos = grub_term_getxy (term);
+	    linux_params.video_cursor_x = pos.x;
+	    linux_params.video_cursor_y = pos.y;
 	    linux_params.video_width = grub_term_width (term);
 	    linux_params.video_height = grub_term_height (term);
 	    found = 1;
@@ -611,8 +594,8 @@ grub_linux_boot (void)
   }
   efi_mmap_buf = (grub_uint8_t *) real_mode_mem + ctx.real_size;
 
-  grub_dprintf ("linux", "real_mode_mem = %lx\n",
-                (unsigned long) real_mode_mem);
+  grub_dprintf ("linux", "real_mode_mem = %p\n",
+                real_mode_mem);
 
   ctx.params = real_mode_mem;
 
@@ -724,7 +707,7 @@ grub_cmd_linux (grub_command_t cmd __attribute__ ((unused)),
       goto fail;
     }
 
-  if (lh.boot_flag != grub_cpu_to_le16 (0xaa55))
+  if (lh.boot_flag != grub_cpu_to_le16_compile_time (0xaa55))
     {
       grub_error (GRUB_ERR_BAD_OS, "invalid magic number");
       goto fail;
@@ -738,7 +721,7 @@ grub_cmd_linux (grub_command_t cmd __attribute__ ((unused)),
 
   /* FIXME: 2.03 is not always good enough (Linux 2.4 can be 2.03 and
      still not support 32-bit boot.  */
-  if (lh.header != grub_cpu_to_le32 (GRUB_LINUX_MAGIC_SIGNATURE)
+  if (lh.header != grub_cpu_to_le32_compile_time (GRUB_LINUX_MAGIC_SIGNATURE)
       || grub_le_to_cpu16 (lh.version) < 0x0203)
     {
       grub_error (GRUB_ERR_BAD_OS, "version too old for 32-bit boot"
@@ -869,7 +852,7 @@ grub_cmd_linux (grub_command_t cmd __attribute__ ((unused)),
   if (grub_le_to_cpu16 (linux_params.version) >= 0x0208)
     {
       linux_params.v0208.efi_signature = GRUB_LINUX_EFI_SIGNATURE;
-      linux_params.v0208.efi_system_table = (grub_uint32_t) (unsigned long) grub_efi_system_table;
+      linux_params.v0208.efi_system_table = (grub_uint32_t) (grub_addr_t) grub_efi_system_table;
 #ifdef __x86_64__
       linux_params.v0208.efi_system_table_hi = (grub_uint32_t) ((grub_uint64_t) grub_efi_system_table >> 32);
 #endif
@@ -877,12 +860,12 @@ grub_cmd_linux (grub_command_t cmd __attribute__ ((unused)),
   else if (grub_le_to_cpu16 (linux_params.version) >= 0x0206)
     {
       linux_params.v0206.efi_signature = GRUB_LINUX_EFI_SIGNATURE;
-      linux_params.v0206.efi_system_table = (grub_uint32_t) (unsigned long) grub_efi_system_table;
+      linux_params.v0206.efi_system_table = (grub_uint32_t) (grub_addr_t) grub_efi_system_table;
     }
   else if (grub_le_to_cpu16 (linux_params.version) >= 0x0204)
     {
       linux_params.v0204.efi_signature = GRUB_LINUX_EFI_SIGNATURE_0204;
-      linux_params.v0204.efi_system_table = (grub_uint32_t) (unsigned long) grub_efi_system_table;
+      linux_params.v0204.efi_system_table = (grub_uint32_t) (grub_addr_t) grub_efi_system_table;
     }
 #endif
 
@@ -932,8 +915,8 @@ grub_cmd_linux (grub_command_t cmd __attribute__ ((unused)),
 	    grub_env_set ("gfxpayload", "text");
 	    grub_printf_ (N_("%s is deprecated. "
 			     "Use set gfxpayload=%s before "
-			     "linux command instead.\n"), "text",
-			  argv[i]);
+			     "linux command instead.\n"),
+			  argv[i], "text");
 	    break;
 
 	  case 1:
@@ -942,8 +925,8 @@ grub_cmd_linux (grub_command_t cmd __attribute__ ((unused)),
 	    grub_env_set ("gfxpayload", "text");
 	    grub_printf_ (N_("%s is deprecated. "
 			     "Use set gfxpayload=%s before "
-			     "linux command instead.\n"), "text",
-			  argv[i]);
+			     "linux command instead.\n"),
+			  argv[i], "text");
 	    break;
 	  default:
 	    /* Ignore invalid values.  */
@@ -1001,10 +984,13 @@ grub_cmd_linux (grub_command_t cmd __attribute__ ((unused)),
 	      {
 	      case 'g':
 		shift += 10;
+		/* FALLTHROUGH */
 	      case 'm':
 		shift += 10;
+		/* FALLTHROUGH */
 	      case 'k':
 		shift += 10;
+		/* FALLTHROUGH */
 	      default:
 		break;
 	      }
@@ -1062,11 +1048,11 @@ static grub_err_t
 grub_cmd_initrd (grub_command_t cmd __attribute__ ((unused)),
 		 int argc, char *argv[])
 {
-  grub_size_t size = 0;
+  grub_size_t size = 0, aligned_size = 0;
   grub_addr_t addr_min, addr_max;
   grub_addr_t addr;
   grub_err_t err;
-  struct grub_linux_initrd_context initrd_ctx;
+  struct grub_linux_initrd_context initrd_ctx = { 0, 0, 0 };
 
   if (argc == 0)
     {
@@ -1084,8 +1070,7 @@ grub_cmd_initrd (grub_command_t cmd __attribute__ ((unused)),
     goto fail;
 
   size = grub_get_initrd_size (&initrd_ctx);
-
-  initrd_pages = (page_align (size) >> 12);
+  aligned_size = ALIGN_UP (size, 4096);
 
   /* Get the highest address available for the initrd.  */
   if (grub_le_to_cpu16 (linux_params.version) >= 0x0203)
@@ -1113,7 +1098,7 @@ grub_cmd_initrd (grub_command_t cmd __attribute__ ((unused)),
   addr_min = (grub_addr_t) prot_mode_target + prot_init_space;
 
   /* Put the initrd as high as possible, 4KiB aligned.  */
-  addr = (addr_max - size) & ~0xFFF;
+  addr = (addr_max - aligned_size) & ~0xFFF;
 
   if (addr < addr_min)
     {
@@ -1124,7 +1109,8 @@ grub_cmd_initrd (grub_command_t cmd __attribute__ ((unused)),
   {
     grub_relocator_chunk_t ch;
     err = grub_relocator_alloc_chunk_align (relocator, &ch,
-					    addr_min, addr, size, 0x1000,
+					    addr_min, addr, aligned_size,
+					    0x1000,
 					    GRUB_RELOCATOR_PREFERENCE_HIGH,
 					    1);
     if (err)
