@@ -51,6 +51,11 @@ struct per_term_screen
   struct grub_term_output *term;
   int y_line_start;
   struct grub_term_screen_geometry geo;
+  /* Scratch variables used when updating. Having them here avoids
+     loads of small mallocs.  */
+  int orig_num;
+  int down;
+  enum update_mode mode;
 };
 
 struct screen
@@ -121,9 +126,13 @@ ensure_space (struct line *linep, int extra)
 static int
 get_logical_num_lines (struct line *linep, struct per_term_screen *term_screen)
 {
-  return (grub_getstringwidth (linep->buf, linep->buf + linep->len,
-			       term_screen->term)
-	  / term_screen->geo.entry_width) + 1;
+  grub_size_t width = grub_getstringwidth (linep->buf, linep->buf + linep->len,
+					   term_screen->term);
+
+  /* Empty line still consumes space on screen */
+  return width ? (width + (unsigned) term_screen->geo.entry_width - 1) /
+		 (unsigned) term_screen->geo.entry_width
+	       : 1;
 }
 
 static void
@@ -146,8 +155,8 @@ print_empty_line (int y, struct per_term_screen *term_screen)
   int i;
 
   grub_term_gotoxy (term_screen->term,
-		    term_screen->geo.first_entry_x,
-		    y + term_screen->geo.first_entry_y);
+		    (struct grub_term_coordinate) { term_screen->geo.first_entry_x,
+			y + term_screen->geo.first_entry_y });
 
   for (i = 0; i < term_screen->geo.entry_width + 1; i++)
     grub_putcode (' ', term_screen->term);
@@ -156,10 +165,11 @@ print_empty_line (int y, struct per_term_screen *term_screen)
 static void
 print_updown (int upflag, int downflag, struct per_term_screen *term_screen)
 {
-  grub_term_gotoxy (term_screen->term, term_screen->geo.first_entry_x
-		    + term_screen->geo.entry_width + 1
-		    + term_screen->geo.border,
-		    term_screen->geo.first_entry_y);
+  grub_term_gotoxy (term_screen->term,
+		    (struct grub_term_coordinate) { term_screen->geo.first_entry_x
+			+ term_screen->geo.entry_width + 1
+			+ term_screen->geo.border,
+			term_screen->geo.first_entry_y });
 
   if (upflag && downflag)
     grub_putcode (GRUB_UNICODE_UPDOWNARROW, term_screen->term);
@@ -175,10 +185,11 @@ print_updown (int upflag, int downflag, struct per_term_screen *term_screen)
 static void
 print_up (int flag, struct per_term_screen *term_screen)
 {
-  grub_term_gotoxy (term_screen->term, term_screen->geo.first_entry_x
-		    + term_screen->geo.entry_width + 1
-		    + term_screen->geo.border,
-		    term_screen->geo.first_entry_y);
+  grub_term_gotoxy (term_screen->term,
+		    (struct grub_term_coordinate) { term_screen->geo.first_entry_x
+			+ term_screen->geo.entry_width + 1
+			+ term_screen->geo.border,
+			term_screen->geo.first_entry_y });
 
   if (flag)
     grub_putcode (GRUB_UNICODE_UPARROW, term_screen->term);
@@ -190,11 +201,12 @@ print_up (int flag, struct per_term_screen *term_screen)
 static void
 print_down (int flag, struct per_term_screen *term_screen)
 {
-  grub_term_gotoxy (term_screen->term, term_screen->geo.first_entry_x
-		    + term_screen->geo.entry_width + 1
-		    + term_screen->geo.border,
-		    term_screen->geo.first_entry_y
-		    + term_screen->geo.num_entries - 1);
+  grub_term_gotoxy (term_screen->term,
+		    (struct grub_term_coordinate) { term_screen->geo.first_entry_x
+			+ term_screen->geo.entry_width + 1
+			+ term_screen->geo.border,
+			term_screen->geo.first_entry_y
+			+ term_screen->geo.num_entries - 1 });
 
   if (flag)
     grub_putcode (GRUB_UNICODE_DOWNARROW, term_screen->term);
@@ -220,10 +232,10 @@ update_screen (struct screen *screen, struct per_term_screen *term_screen,
   for (i = 0; i < screen->line; i++, linep++)
     y += get_logical_num_lines (linep, term_screen);
   linep = screen->lines + screen->line;
-  int t = grub_getstringwidth (linep->buf, linep->buf + screen->column,
-			       term_screen->term);
-  y += t / term_screen->geo.entry_width;
-  if (t % term_screen->geo.entry_width == 0
+  grub_size_t t = grub_getstringwidth (linep->buf, linep->buf + screen->column,
+				       term_screen->term);
+  y += t / (unsigned) term_screen->geo.entry_width;
+  if (t % (unsigned) term_screen->geo.entry_width == 0
       && t != 0 &&  screen->column == linep->len)
     y--;
   /* Check if scrolling is necessary.  */
@@ -242,6 +254,8 @@ update_screen (struct screen *screen, struct per_term_screen *term_screen,
       mode = ALL_LINES;
     }
 
+  grub_term_setcursor (term_screen->term, 0);
+
   if (mode != NO_LINE)
     {
       /* Draw lines. This code is tricky, because this must calculate logical
@@ -278,9 +292,10 @@ update_screen (struct screen *screen, struct per_term_screen *term_screen,
 	  if (i == region_start || linep == screen->lines + screen->line
 	      || (i > region_start && mode == ALL_LINES))
 	    {
-	      grub_term_gotoxy (term_screen->term, term_screen->geo.first_entry_x,
-				(y < 0 ? 0 : y)
-				+ term_screen->geo.first_entry_y);
+	      grub_term_gotoxy (term_screen->term,
+				(struct grub_term_coordinate) { term_screen->geo.first_entry_x,
+				    (y < 0 ? 0 : y)
+				    + term_screen->geo.first_entry_y });
 
 	      grub_print_ucs4_menu (linep->buf,
 				    linep->buf + linep->len,
@@ -338,16 +353,18 @@ update_screen (struct screen *screen, struct per_term_screen *term_screen,
 	y += get_logical_num_lines (screen->lines + i, term_screen);
       if (cpos >= &(screen->lines[screen->line].pos[term_screen - screen->terms])[0])
 	grub_term_gotoxy (term_screen->term, 
-			  cpos->x + term_screen->geo.first_entry_x,
-			  cpos->y + y
-			  + term_screen->geo.first_entry_y);
+			  (struct grub_term_coordinate) { cpos->x + term_screen->geo.first_entry_x,
+			      cpos->y + y
+			      + term_screen->geo.first_entry_y });
       else
 	grub_term_gotoxy (term_screen->term, 
-			  term_screen->geo.first_entry_x,
-			  y + term_screen->geo.first_entry_y);
+			  (struct grub_term_coordinate) { term_screen->geo.first_entry_x,
+			      y + term_screen->geo.first_entry_y });
 
     }
 
+  grub_term_setcursor (term_screen->term, 1);
+
   grub_term_refresh (term_screen->term);
 }
 
@@ -367,14 +384,12 @@ insert_string (struct screen *screen, const char *s, int update)
 {
   int region_start = screen->num_lines;
   int region_column = 0;
-  int down[screen->nterms];
-  enum update_mode mode[screen->nterms];
   unsigned i;
 
   for (i = 0; i < screen->nterms; i++)
     {
-      down[i] = 0;
-      mode[i] = NO_LINE;
+      screen->terms[i].down = 0;
+      screen->terms[i].mode = NO_LINE;
     }
 
   while (*s)
@@ -432,8 +447,8 @@ insert_string (struct screen *screen, const char *s, int update)
 
 	  for (i = 0; i < screen->nterms; i++)
 	    {
-	      mode[i] = ALL_LINES;
-	      down[i] = 1; /* XXX not optimal.  */
+	      screen->terms[i].mode = ALL_LINES;
+	      screen->terms[i].down = 1; /* XXX not optimal.  */
 	    }
 
 	  /* Move the cursor.  */
@@ -447,7 +462,6 @@ insert_string (struct screen *screen, const char *s, int update)
 	  const char *p;
 	  struct line *current_linep;
 	  int size;
-	  int orig_num[screen->nterms], new_num[screen->nterms];
 	  grub_uint32_t *unicode_msg;
 
 	  /* Find a string delimited by LF.  */
@@ -466,7 +480,10 @@ insert_string (struct screen *screen, const char *s, int update)
 				    (grub_uint8_t *) s, (p - s), 0);
 
 	  if (! ensure_space (current_linep, size))
-	    return 0;
+	    {
+	      grub_free (unicode_msg);
+	      return 0;
+	    }
 
 	  grub_memmove (current_linep->buf + screen->column + size,
 			current_linep->buf + screen->column,
@@ -484,12 +501,9 @@ insert_string (struct screen *screen, const char *s, int update)
 	    }
 
 	  for (i = 0; i < screen->nterms; i++)
-	    orig_num[i] = get_logical_num_lines (current_linep,
+	    screen->terms[i].orig_num = get_logical_num_lines (current_linep,
 						 &screen->terms[i]);
 	  current_linep->len += size;
-	  for (i = 0; i < screen->nterms; i++)
-	    new_num[i] = get_logical_num_lines (current_linep,
-						&screen->terms[i]);
 
 	  /* Update the dirty region.  */
 	  if (region_start > screen->line)
@@ -499,13 +513,17 @@ insert_string (struct screen *screen, const char *s, int update)
 	    }
 
 	  for (i = 0; i < screen->nterms; i++)
-	    if (orig_num[i] != new_num[i])
-	      {
-		mode[i] = ALL_LINES;
-		down[i] = 1; /* XXX not optimal.  */
-	      }
-	    else if (mode[i] != ALL_LINES)
-	      mode[i] = SINGLE_LINE;
+	    {
+	      int new_num = get_logical_num_lines (current_linep,
+						   &screen->terms[i]);
+	      if (screen->terms[i].orig_num != new_num)
+		{
+		  screen->terms[i].mode = ALL_LINES;
+		  screen->terms[i].down = 1; /* XXX not optimal.  */
+		}
+	      else if (screen->terms[i].mode != ALL_LINES)
+		screen->terms[i].mode = SINGLE_LINE;
+	    }
 
 	  /* Move the cursor.  */
 	  advance_to (screen, screen->column + size);
@@ -518,7 +536,8 @@ insert_string (struct screen *screen, const char *s, int update)
   if (update)
     for (i = 0; i < screen->nterms; i++)
       update_screen (screen, &screen->terms[i],
-		     region_start, region_column, 0, down[i], mode[i]);
+		     region_start, region_column, 0, screen->terms[i].down,
+		     screen->terms[i].mode);
 
   return 1;
 }
@@ -750,11 +769,10 @@ delete_char (struct screen *screen, int update)
   linep = screen->lines + screen->line;
   if (linep->len > screen->column)
     {
-      int orig_num[screen->nterms], new_num;
       unsigned i;
 
       for (i = 0; i < screen->nterms; i++)
-	orig_num[i] = get_logical_num_lines (linep, &screen->terms[i]);
+	screen->terms[i].orig_num = get_logical_num_lines (linep, &screen->terms[i]);
 
       grub_memmove (linep->buf + screen->column,
 		    linep->buf + screen->column + 1,
@@ -776,8 +794,9 @@ delete_char (struct screen *screen, int update)
 	{
 	  for (i = 0; i < screen->nterms; i++)
 	    {
+	      int new_num;
 	      new_num = get_logical_num_lines (linep, &screen->terms[i]);
-	      if (orig_num[i] != new_num)
+	      if (screen->terms[i].orig_num != new_num)
 		update_screen (screen, &screen->terms[i],
 			       start, column, 0, 0, ALL_LINES);
 	      else
@@ -863,7 +882,6 @@ kill_line (struct screen *screen, int continuous, int update)
 
   if (size > 0)
     {
-      int orig_num[screen->nterms], new_num;
       unsigned i;
 
       p = grub_realloc (p, offset + size + 1);
@@ -876,15 +894,16 @@ kill_line (struct screen *screen, int continuous, int update)
       screen->killed_text = p;
 
       for (i = 0; i < screen->nterms; i++)
-	orig_num[i] = get_logical_num_lines (linep, &screen->terms[i]);
+	screen->terms[i].orig_num = get_logical_num_lines (linep, &screen->terms[i]);
       linep->len = screen->column;
 
       if (update)
 	{
 	  for (i = 0; i < screen->nterms; i++)
 	    {
+	      int new_num;
 	      new_num = get_logical_num_lines (linep, &screen->terms[i]);
-	      if (orig_num[i] != new_num)
+	      if (screen->terms[i].orig_num != new_num)
 		update_screen (screen, &screen->terms[i],
 			       screen->line, screen->column, 0, 1, ALL_LINES);
 	      else
@@ -1021,22 +1040,25 @@ complete (struct screen *screen, int continuous, int update)
       if (restore)
 	for (i = 0; i < screen->nterms; i++)
 	  {
-	    int width = grub_term_width (screen->terms[i].term) - 2;
+	    unsigned width = grub_term_width (screen->terms[i].term);
+	    if (width > 2)
+	      width -= 2;
 	    if (width > 15)
 	      width -= 6;
-	    int num_sections = ((completion_buffer.len
-				 + width - 1)
-				/ width);
+	    unsigned num_sections = ((completion_buffer.len
+				      + width - 1)
+				     / width);
 	    grub_uint32_t *endp;
-	    grub_uint16_t pos;
+	    struct grub_term_coordinate pos;
 	    grub_uint32_t *p = ucs4;
 
 	    pos = grub_term_getxy (screen->terms[i].term);
 
 	    screen->completion_shown = 1;
 
-	    grub_term_gotoxy (screen->terms[i].term, 0,
-			      screen->terms[i].geo.timeout_y);
+	    grub_term_gotoxy (screen->terms[i].term,
+			      (struct grub_term_coordinate) { 0,
+				  screen->terms[i].geo.timeout_y });
 	    if (screen->terms[i].geo.timeout_lines >= 2)
 	      {
 		grub_puts_terminal ("   ", screen->terms[i].term);
@@ -1071,7 +1093,7 @@ complete (struct screen *screen, int continuous, int update)
 		grub_puts_terminal ("\n    ", screen->terms[i].term);
 	      }
 
-	    p += (count % num_sections) * width;
+	    p += ((unsigned) count % num_sections) * width;
 	    endp = p + width;
 
 	    if (p != ucs4)
@@ -1084,7 +1106,7 @@ complete (struct screen *screen, int continuous, int update)
 
 	    if (ucs4 + ucs4len > endp)
 	      grub_putcode (GRUB_UNICODE_RIGHTARROW, screen->terms[i].term);
-	    grub_term_gotoxy (screen->terms[i].term, pos >> 8, pos & 0xFF);
+	    grub_term_gotoxy (screen->terms[i].term, pos);
 	  }
     }
 
@@ -1105,13 +1127,14 @@ complete (struct screen *screen, int continuous, int update)
 static void
 clear_completions (struct per_term_screen *term_screen)
 {
-  grub_uint16_t pos;
+  struct grub_term_coordinate pos;
   unsigned j;
   int i;
 
   pos = grub_term_getxy (term_screen->term);
-  grub_term_gotoxy (term_screen->term, 0,
-		    term_screen->geo.timeout_y);
+  grub_term_gotoxy (term_screen->term,
+		    (struct grub_term_coordinate) { 0,
+			term_screen->geo.timeout_y });
 
   for (i = 0; i < term_screen->geo.timeout_lines; i++)
     {
@@ -1121,7 +1144,7 @@ clear_completions (struct per_term_screen *term_screen)
 	grub_putcode ('\n', term_screen->term);
     }
 
-  grub_term_gotoxy (term_screen->term, pos >> 8, pos & 0xFF);
+  grub_term_gotoxy (term_screen->term, pos);
   grub_term_refresh (term_screen->term);
 }
 
@@ -1181,7 +1204,7 @@ run (struct screen *screen)
       }
     script[size] = '\0';
   }
-  grub_script_execute_sourcecode (script, 0, dummy);
+  grub_script_execute_new_scope (script, 0, dummy);
   grub_free (script);
 
   if (errs_before != grub_err_printed_errors)
@@ -1249,6 +1272,7 @@ grub_menu_entry_run (grub_menu_entry_t entry)
       if (! screen->lines[i].pos)
 	{
 	  grub_print_error ();
+	  destroy_screen (screen);
 	  grub_errno = GRUB_ERR_NONE;
 	  return;
 	}
@@ -1258,6 +1282,7 @@ grub_menu_entry_run (grub_menu_entry_t entry)
   if (!screen->terms)
     {
       grub_print_error ();
+      destroy_screen (screen);
       grub_errno = GRUB_ERR_NONE;
       return;
     }
@@ -1384,7 +1409,7 @@ grub_menu_entry_run (grub_menu_entry_t entry)
 
 	case GRUB_TERM_CTRL | 'c':
 	case GRUB_TERM_KEY_F2:
-	  grub_cmdline_run (1);
+	  grub_cmdline_run (1, 0);
 	  goto refresh;
 
 	case GRUB_TERM_CTRL | 'x':
