@@ -49,6 +49,9 @@
 #define GRUB_EFI_MEMORY_WP	0x0000000000001000LL
 #define GRUB_EFI_MEMORY_RP	0x0000000000002000LL
 #define GRUB_EFI_MEMORY_XP	0x0000000000004000LL
+#define GRUB_EFI_MEMORY_NV	0x0000000000008000LL
+#define GRUB_EFI_MEMORY_MORE_RELIABLE	0x0000000000010000LL
+#define GRUB_EFI_MEMORY_RO	0x0000000000020000LL
 #define GRUB_EFI_MEMORY_RUNTIME	0x8000000000000000LL
 
 #define GRUB_EFI_OPEN_PROTOCOL_BY_HANDLE_PROTOCOL	0x00000001
@@ -111,7 +114,7 @@
     { 0x8e, 0x39, 0x00, 0xa0, 0xc9, 0x69, 0x72, 0x3b } \
   }
 
-#define EFI_SIMPLE_TEXT_INPUT_EX_PROTOCOL_GUID \
+#define GRUB_EFI_SIMPLE_TEXT_INPUT_EX_PROTOCOL_GUID \
   { 0xdd9e7534, 0x7762, 0x4698, \
     { 0x8c, 0x14, 0xf5, 0x85, 0x17, 0xa6, 0x25, 0xaa } \
   }
@@ -246,6 +249,51 @@
     { 0x8E, 0x8B, 0xBB, 0xA2, 0x0B, 0x1B, 0x5B, 0x75 } \
   }
 
+#define GRUB_EFI_TIANO_CUSTOM_DECOMPRESS_GUID \
+  { 0xa31280ad, 0x481e, 0x41b6, \
+    { 0x95, 0xe8, 0x12, 0x7f, 0x4c, 0x98, 0x47, 0x79 } \
+  }
+
+#define GRUB_EFI_CRC32_GUIDED_SECTION_EXTRACTION_GUID \
+  { 0xfc1bcdb0, 0x7d31, 0x49aa, \
+    { 0x93, 0x6a, 0xa4, 0x60, 0x0d, 0x9d, 0xd0, 0x83 } \
+  }
+
+#define GRUB_EFI_LZMA_CUSTOM_DECOMPRESS_GUID \
+  { 0xee4e5898, 0x3914, 0x4259, \
+    { 0x9d, 0x6e, 0xdc, 0x7b, 0xd7, 0x94, 0x03, 0xcf } \
+  }
+
+#define GRUB_EFI_TSC_FREQUENCY_GUID \
+  { 0xdba6a7e3, 0xbb57, 0x4be7, \
+    { 0x8a, 0xf8, 0xd5, 0x78, 0xdb, 0x7e, 0x56, 0x87 } \
+  }
+
+#define GRUB_EFI_SYSTEM_RESOURCE_TABLE_GUID \
+  { 0xb122a263, 0x3661, 0x4f68, \
+    { 0x99, 0x29, 0x78, 0xf8, 0xb0, 0xd6, 0x21, 0x80 } \
+  }
+
+#define GRUB_EFI_DXE_SERVICES_TABLE_GUID \
+  { 0x05ad34ba, 0x6f02, 0x4214, \
+    { 0x95, 0x2e, 0x4d, 0xa0, 0x39, 0x8e, 0x2b, 0xb9 } \
+  }
+
+#define GRUB_EFI_HOB_LIST_GUID \
+  { 0x7739f24c, 0x93d7, 0x11d4, \
+    { 0x9a, 0x3a, 0x00, 0x90, 0x27, 0x3f, 0xc1, 0x4d } \
+  }
+
+#define GRUB_EFI_MEMORY_TYPE_INFORMATION_GUID \
+  { 0x4c19049f, 0x4137, 0x4dd3, \
+    { 0x9c, 0x10, 0x8b, 0x97, 0xa8, 0x3f, 0xfd, 0xfa } \
+  }
+
+#define GRUB_EFI_DEBUG_IMAGE_INFO_TABLE_GUID \
+  { 0x49152e77, 0x1ada, 0x4764, \
+    { 0xb7, 0xa2, 0x7a, 0xfe, 0xfe, 0xd9, 0x5e, 0x8b } \
+  }
+
 #define GRUB_EFI_MPS_TABLE_GUID	\
   { 0xeb9d2d2f, 0x2d88, 0x11d3, \
     { 0x9a, 0x16, 0x00, 0x90, 0x27, 0x3f, 0xc1, 0x4d } \
@@ -276,6 +324,16 @@
       { 0x82, 0x79, 0xa8, 0x4b, 0x79, 0x61, 0x78, 0x98 } \
   }
 
+#define GRUB_EFI_DEVICE_TREE_GUID \
+  { 0xb1b621d5, 0xf19c, 0x41a5, \
+      { 0x83, 0x0b, 0xd9, 0x15, 0x2c, 0x69, 0xaa, 0xe0 } \
+  }
+
+#define GRUB_EFI_VENDOR_APPLE_GUID \
+  { 0x2B0585EB, 0xD8B8, 0x49A9,	\
+      { 0x8B, 0x8C, 0xE2, 0x1B, 0x01, 0xAE, 0xF2, 0xB7 } \
+  }
+
 struct grub_efi_sal_system_table
 {
   grub_uint32_t signature;
@@ -421,6 +479,7 @@ enum grub_efi_memory_type
     GRUB_EFI_MEMORY_MAPPED_IO,
     GRUB_EFI_MEMORY_MAPPED_IO_PORT_SPACE,
     GRUB_EFI_PAL_CODE,
+    GRUB_EFI_PERSISTENT_MEMORY,
     GRUB_EFI_MAX_MEMORY_TYPE
   };
 typedef enum grub_efi_memory_type grub_efi_memory_type_t;
@@ -449,8 +508,13 @@ typedef enum grub_efi_reset_type grub_efi_reset_type_t;
 
 /* Types.  */
 typedef char grub_efi_boolean_t;
-typedef long grub_efi_intn_t;
-typedef unsigned long grub_efi_uintn_t;
+#if GRUB_CPU_SIZEOF_VOID_P == 8
+typedef grub_int64_t grub_efi_intn_t;
+typedef grub_uint64_t grub_efi_uintn_t;
+#else
+typedef grub_int32_t grub_efi_intn_t;
+typedef grub_uint32_t grub_efi_uintn_t;
+#endif
 typedef grub_int8_t grub_efi_int8_t;
 typedef grub_uint8_t grub_efi_uint8_t;
 typedef grub_int16_t grub_efi_int16_t;
@@ -462,12 +526,10 @@ typedef grub_uint64_t grub_efi_uint64_t;
 typedef grub_uint8_t grub_efi_char8_t;
 typedef grub_uint16_t grub_efi_char16_t;
 
-#define PRIxGRUB_EFI_UINTN_T "lx"
-
 typedef grub_efi_intn_t grub_efi_status_t;
 
 #define GRUB_EFI_ERROR_CODE(value)	\
-  ((1L << (sizeof (grub_efi_status_t) * 8 - 1)) | (value))
+  ((((grub_efi_status_t) 1) << (sizeof (grub_efi_status_t) * 8 - 1)) | (value))
 
 #define GRUB_EFI_WARNING_CODE(value)	(value)
 
@@ -526,6 +588,15 @@ struct grub_efi_guid
 } __attribute__ ((aligned(8)));
 typedef struct grub_efi_guid grub_efi_guid_t;
 
+struct grub_efi_packed_guid
+{
+  grub_uint32_t data1;
+  grub_uint16_t data2;
+  grub_uint16_t data3;
+  grub_uint8_t data4[8];
+} GRUB_PACKED;
+typedef struct grub_efi_packed_guid grub_efi_packed_guid_t;
+
 /* XXX although the spec does not specify the padding, this actually
    must have the padding!  */
 struct grub_efi_memory_descriptor
@@ -536,7 +607,7 @@ struct grub_efi_memory_descriptor
   grub_efi_virtual_address_t virtual_start;
   grub_efi_uint64_t num_pages;
   grub_efi_uint64_t attribute;
-} __attribute__ ((packed));
+} GRUB_PACKED;
 typedef struct grub_efi_memory_descriptor grub_efi_memory_descriptor_t;
 
 /* Device Path definitions.  */
@@ -544,8 +615,8 @@ struct grub_efi_device_path
 {
   grub_efi_uint8_t type;
   grub_efi_uint8_t subtype;
-  grub_efi_uint8_t length[2];
-};
+  grub_efi_uint16_t length;
+} GRUB_PACKED;
 typedef struct grub_efi_device_path grub_efi_device_path_t;
 /* XXX EFI does not define EFI_DEVICE_PATH_PROTOCOL but uses it.
    It seems to be identical to EFI_DEVICE_PATH.  */
@@ -553,7 +624,7 @@ typedef struct grub_efi_device_path grub_efi_device_path_protocol_t;
 
 #define GRUB_EFI_DEVICE_PATH_TYPE(dp)		((dp)->type & 0x7f)
 #define GRUB_EFI_DEVICE_PATH_SUBTYPE(dp)	((dp)->subtype)
-#define GRUB_EFI_DEVICE_PATH_LENGTH(dp)		(grub_get_unaligned16 ((dp)->length))
+#define GRUB_EFI_DEVICE_PATH_LENGTH(dp)		((dp)->length)
 
 /* The End of Device Path nodes.  */
 #define GRUB_EFI_END_DEVICE_PATH_TYPE			(0xff & 0x7f)
@@ -580,7 +651,7 @@ struct grub_efi_pci_device_path
   grub_efi_device_path_t header;
   grub_efi_uint8_t function;
   grub_efi_uint8_t device;
-} __attribute__ ((packed));
+} GRUB_PACKED;
 typedef struct grub_efi_pci_device_path grub_efi_pci_device_path_t;
 
 #define GRUB_EFI_PCCARD_DEVICE_PATH_SUBTYPE		2
@@ -589,7 +660,7 @@ struct grub_efi_pccard_device_path
 {
   grub_efi_device_path_t header;
   grub_efi_uint8_t function;
-} __attribute__ ((packed));
+} GRUB_PACKED;
 typedef struct grub_efi_pccard_device_path grub_efi_pccard_device_path_t;
 
 #define GRUB_EFI_MEMORY_MAPPED_DEVICE_PATH_SUBTYPE	3
@@ -600,7 +671,7 @@ struct grub_efi_memory_mapped_device_path
   grub_efi_uint32_t memory_type;
   grub_efi_physical_address_t start_address;
   grub_efi_physical_address_t end_address;
-} __attribute__ ((packed));
+} GRUB_PACKED;
 typedef struct grub_efi_memory_mapped_device_path grub_efi_memory_mapped_device_path_t;
 
 #define GRUB_EFI_VENDOR_DEVICE_PATH_SUBTYPE		4
@@ -608,9 +679,9 @@ typedef struct grub_efi_memory_mapped_device_path grub_efi_memory_mapped_device_
 struct grub_efi_vendor_device_path
 {
   grub_efi_device_path_t header;
-  grub_efi_guid_t vendor_guid;
+  grub_efi_packed_guid_t vendor_guid;
   grub_efi_uint8_t vendor_defined_data[0];
-} __attribute__ ((packed));
+} GRUB_PACKED;
 typedef struct grub_efi_vendor_device_path grub_efi_vendor_device_path_t;
 
 #define GRUB_EFI_CONTROLLER_DEVICE_PATH_SUBTYPE		5
@@ -619,7 +690,7 @@ struct grub_efi_controller_device_path
 {
   grub_efi_device_path_t header;
   grub_efi_uint32_t controller_number;
-} __attribute__ ((packed));
+} GRUB_PACKED;
 typedef struct grub_efi_controller_device_path grub_efi_controller_device_path_t;
 
 /* ACPI Device Path.  */
@@ -632,7 +703,7 @@ struct grub_efi_acpi_device_path
   grub_efi_device_path_t header;
   grub_efi_uint32_t hid;
   grub_efi_uint32_t uid;
-} __attribute__ ((packed));
+} GRUB_PACKED;
 typedef struct grub_efi_acpi_device_path grub_efi_acpi_device_path_t;
 
 #define GRUB_EFI_EXPANDED_ACPI_DEVICE_PATH_SUBTYPE	2
@@ -644,7 +715,7 @@ struct grub_efi_expanded_acpi_device_path
   grub_efi_uint32_t uid;
   grub_efi_uint32_t cid;
   char hidstr[0];
-} __attribute__ ((packed));
+} GRUB_PACKED;
 typedef struct grub_efi_expanded_acpi_device_path grub_efi_expanded_acpi_device_path_t;
 
 #define GRUB_EFI_EXPANDED_ACPI_HIDSTR(dp)	\
@@ -667,7 +738,7 @@ struct grub_efi_atapi_device_path
   grub_efi_uint8_t primary_secondary;
   grub_efi_uint8_t slave_master;
   grub_efi_uint16_t lun;
-} __attribute__ ((packed));
+} GRUB_PACKED;
 typedef struct grub_efi_atapi_device_path grub_efi_atapi_device_path_t;
 
 #define GRUB_EFI_SCSI_DEVICE_PATH_SUBTYPE		2
@@ -677,7 +748,7 @@ struct grub_efi_scsi_device_path
   grub_efi_device_path_t header;
   grub_efi_uint16_t pun;
   grub_efi_uint16_t lun;
-} __attribute__ ((packed));
+} GRUB_PACKED;
 typedef struct grub_efi_scsi_device_path grub_efi_scsi_device_path_t;
 
 #define GRUB_EFI_FIBRE_CHANNEL_DEVICE_PATH_SUBTYPE	3
@@ -688,7 +759,7 @@ struct grub_efi_fibre_channel_device_path
   grub_efi_uint32_t reserved;
   grub_efi_uint64_t wwn;
   grub_efi_uint64_t lun;
-} __attribute__ ((packed));
+} GRUB_PACKED;
 typedef struct grub_efi_fibre_channel_device_path grub_efi_fibre_channel_device_path_t;
 
 #define GRUB_EFI_1394_DEVICE_PATH_SUBTYPE		4
@@ -698,7 +769,7 @@ struct grub_efi_1394_device_path
   grub_efi_device_path_t header;
   grub_efi_uint32_t reserved;
   grub_efi_uint64_t guid;
-} __attribute__ ((packed));
+} GRUB_PACKED;
 typedef struct grub_efi_1394_device_path grub_efi_1394_device_path_t;
 
 #define GRUB_EFI_USB_DEVICE_PATH_SUBTYPE		5
@@ -707,8 +778,8 @@ struct grub_efi_usb_device_path
 {
   grub_efi_device_path_t header;
   grub_efi_uint8_t parent_port_number;
-  grub_efi_uint8_t interface;
-} __attribute__ ((packed));
+  grub_efi_uint8_t usb_interface;
+} GRUB_PACKED;
 typedef struct grub_efi_usb_device_path grub_efi_usb_device_path_t;
 
 #define GRUB_EFI_USB_CLASS_DEVICE_PATH_SUBTYPE		15
@@ -721,7 +792,7 @@ struct grub_efi_usb_class_device_path
   grub_efi_uint8_t device_class;
   grub_efi_uint8_t device_subclass;
   grub_efi_uint8_t device_protocol;
-} __attribute__ ((packed));
+} GRUB_PACKED;
 typedef struct grub_efi_usb_class_device_path grub_efi_usb_class_device_path_t;
 
 #define GRUB_EFI_I2O_DEVICE_PATH_SUBTYPE		6
@@ -730,7 +801,7 @@ struct grub_efi_i2o_device_path
 {
   grub_efi_device_path_t header;
   grub_efi_uint32_t tid;
-} __attribute__ ((packed));
+} GRUB_PACKED;
 typedef struct grub_efi_i2o_device_path grub_efi_i2o_device_path_t;
 
 #define GRUB_EFI_MAC_ADDRESS_DEVICE_PATH_SUBTYPE	11
@@ -740,7 +811,7 @@ struct grub_efi_mac_address_device_path
   grub_efi_device_path_t header;
   grub_efi_mac_address_t mac_address;
   grub_efi_uint8_t if_type;
-} __attribute__ ((packed));
+} GRUB_PACKED;
 typedef struct grub_efi_mac_address_device_path grub_efi_mac_address_device_path_t;
 
 #define GRUB_EFI_IPV4_DEVICE_PATH_SUBTYPE		12
@@ -754,7 +825,7 @@ struct grub_efi_ipv4_device_path
   grub_efi_uint16_t remote_port;
   grub_efi_uint16_t protocol;
   grub_efi_uint8_t static_ip_address;
-} __attribute__ ((packed));
+} GRUB_PACKED;
 typedef struct grub_efi_ipv4_device_path grub_efi_ipv4_device_path_t;
 
 #define GRUB_EFI_IPV6_DEVICE_PATH_SUBTYPE		13
@@ -768,7 +839,7 @@ struct grub_efi_ipv6_device_path
   grub_efi_uint16_t remote_port;
   grub_efi_uint16_t protocol;
   grub_efi_uint8_t static_ip_address;
-} __attribute__ ((packed));
+} GRUB_PACKED;
 typedef struct grub_efi_ipv6_device_path grub_efi_ipv6_device_path_t;
 
 #define GRUB_EFI_INFINIBAND_DEVICE_PATH_SUBTYPE		9
@@ -781,7 +852,7 @@ struct grub_efi_infiniband_device_path
   grub_efi_uint64_t remote_id;
   grub_efi_uint64_t target_port_id;
   grub_efi_uint64_t device_id;
-} __attribute__ ((packed));
+} GRUB_PACKED;
 typedef struct grub_efi_infiniband_device_path grub_efi_infiniband_device_path_t;
 
 #define GRUB_EFI_UART_DEVICE_PATH_SUBTYPE		14
@@ -794,18 +865,21 @@ struct grub_efi_uart_device_path
   grub_efi_uint8_t data_bits;
   grub_efi_uint8_t parity;
   grub_efi_uint8_t stop_bits;
-} __attribute__ ((packed));
+} GRUB_PACKED;
 typedef struct grub_efi_uart_device_path grub_efi_uart_device_path_t;
 
-#define GRUB_EFI_VENDOR_MESSAGING_DEVICE_PATH_SUBTYPE	10
+#define GRUB_EFI_SATA_DEVICE_PATH_SUBTYPE		18
 
-struct grub_efi_vendor_messaging_device_path
+struct grub_efi_sata_device_path
 {
   grub_efi_device_path_t header;
-  grub_efi_guid_t vendor_guid;
-  grub_efi_uint8_t vendor_defined_data[0];
-} __attribute__ ((packed));
-typedef struct grub_efi_vendor_messaging_device_path grub_efi_vendor_messaging_device_path_t;
+  grub_efi_uint16_t hba_port;
+  grub_efi_uint16_t multiplier_port;
+  grub_efi_uint16_t lun;
+} GRUB_PACKED;
+typedef struct grub_efi_sata_device_path grub_efi_sata_device_path_t;
+
+#define GRUB_EFI_VENDOR_MESSAGING_DEVICE_PATH_SUBTYPE	10
 
 /* Media Device Path.  */
 #define GRUB_EFI_MEDIA_DEVICE_PATH_TYPE			4
@@ -818,10 +892,10 @@ struct grub_efi_hard_drive_device_path
   grub_efi_uint32_t partition_number;
   grub_efi_lba_t partition_start;
   grub_efi_lba_t partition_size;
-  grub_efi_uint8_t partition_signature[8];
-  grub_efi_uint8_t mbr_type;
+  grub_efi_uint8_t partition_signature[16];
+  grub_efi_uint8_t partmap_type;
   grub_efi_uint8_t signature_type;
-} __attribute__ ((packed));
+} GRUB_PACKED;
 typedef struct grub_efi_hard_drive_device_path grub_efi_hard_drive_device_path_t;
 
 #define GRUB_EFI_CDROM_DEVICE_PATH_SUBTYPE		2
@@ -832,7 +906,7 @@ struct grub_efi_cdrom_device_path
   grub_efi_uint32_t boot_entry;
   grub_efi_lba_t partition_start;
   grub_efi_lba_t partition_size;
-} __attribute__ ((packed));
+} GRUB_PACKED;
 typedef struct grub_efi_cdrom_device_path grub_efi_cdrom_device_path_t;
 
 #define GRUB_EFI_VENDOR_MEDIA_DEVICE_PATH_SUBTYPE	3
@@ -840,9 +914,9 @@ typedef struct grub_efi_cdrom_device_path grub_efi_cdrom_device_path_t;
 struct grub_efi_vendor_media_device_path
 {
   grub_efi_device_path_t header;
-  grub_efi_guid_t vendor_guid;
+  grub_efi_packed_guid_t vendor_guid;
   grub_efi_uint8_t vendor_defined_data[0];
-} __attribute__ ((packed));
+} GRUB_PACKED;
 typedef struct grub_efi_vendor_media_device_path grub_efi_vendor_media_device_path_t;
 
 #define GRUB_EFI_FILE_PATH_DEVICE_PATH_SUBTYPE		4
@@ -851,7 +925,7 @@ struct grub_efi_file_path_device_path
 {
   grub_efi_device_path_t header;
   grub_efi_char16_t path_name[0];
-} __attribute__ ((packed));
+} GRUB_PACKED;
 typedef struct grub_efi_file_path_device_path grub_efi_file_path_device_path_t;
 
 #define GRUB_EFI_PROTOCOL_DEVICE_PATH_SUBTYPE		5
@@ -859,8 +933,8 @@ typedef struct grub_efi_file_path_device_path grub_efi_file_path_device_path_t;
 struct grub_efi_protocol_device_path
 {
   grub_efi_device_path_t header;
-  grub_efi_guid_t guid;
-} __attribute__ ((packed));
+  grub_efi_packed_guid_t guid;
+} GRUB_PACKED;
 typedef struct grub_efi_protocol_device_path grub_efi_protocol_device_path_t;
 
 #define GRUB_EFI_PIWG_DEVICE_PATH_SUBTYPE		6
@@ -868,8 +942,8 @@ typedef struct grub_efi_protocol_device_path grub_efi_protocol_device_path_t;
 struct grub_efi_piwg_device_path
 {
   grub_efi_device_path_t header;
-  grub_efi_guid_t guid __attribute__ ((packed));
-} __attribute__ ((packed));
+  grub_efi_packed_guid_t guid;
+} GRUB_PACKED;
 typedef struct grub_efi_piwg_device_path grub_efi_piwg_device_path_t;
 
 
@@ -884,7 +958,7 @@ struct grub_efi_bios_device_path
   grub_efi_uint16_t device_type;
   grub_efi_uint16_t status_flags;
   char description[0];
-} __attribute__ ((packed));
+} GRUB_PACKED;
 typedef struct grub_efi_bios_device_path grub_efi_bios_device_path_t;
 
 struct grub_efi_open_protocol_information_entry
@@ -909,7 +983,7 @@ struct grub_efi_time
   grub_efi_int16_t time_zone;
   grub_efi_uint8_t daylight;
   grub_efi_uint8_t pad2;
-} __attribute__ ((packed));
+} GRUB_PACKED;
 typedef struct grub_efi_time grub_efi_time_t;
 
 struct grub_efi_time_capabilities
@@ -927,6 +1001,32 @@ struct grub_efi_input_key
 };
 typedef struct grub_efi_input_key grub_efi_input_key_t;
 
+typedef grub_efi_uint8_t grub_efi_key_toggle_state_t;
+struct grub_efi_key_state
+{
+	grub_efi_uint32_t key_shift_state;
+	grub_efi_key_toggle_state_t key_toggle_state;
+};
+typedef struct grub_efi_key_state grub_efi_key_state_t;
+
+#define GRUB_EFI_SHIFT_STATE_VALID     0x80000000
+#define GRUB_EFI_RIGHT_SHIFT_PRESSED   0x00000001
+#define GRUB_EFI_LEFT_SHIFT_PRESSED    0x00000002
+#define GRUB_EFI_RIGHT_CONTROL_PRESSED 0x00000004
+#define GRUB_EFI_LEFT_CONTROL_PRESSED  0x00000008
+#define GRUB_EFI_RIGHT_ALT_PRESSED     0x00000010
+#define GRUB_EFI_LEFT_ALT_PRESSED      0x00000020
+#define GRUB_EFI_RIGHT_LOGO_PRESSED    0x00000040
+#define GRUB_EFI_LEFT_LOGO_PRESSED     0x00000080
+#define GRUB_EFI_MENU_KEY_PRESSED      0x00000100
+#define GRUB_EFI_SYS_REQ_PRESSED       0x00000200
+
+#define GRUB_EFI_TOGGLE_STATE_VALID 0x80
+#define GRUB_EFI_KEY_STATE_EXPOSED  0x40
+#define GRUB_EFI_SCROLL_LOCK_ACTIVE 0x01
+#define GRUB_EFI_NUM_LOCK_ACTIVE    0x02
+#define GRUB_EFI_CAPS_LOCK_ACTIVE   0x04
+
 struct grub_efi_simple_text_output_mode
 {
   grub_efi_int32_t max_mode;
@@ -1014,8 +1114,8 @@ struct grub_efi_boot_services
    grub_efi_status_t
    (*install_protocol_interface) (grub_efi_handle_t *handle,
 				  grub_efi_guid_t *protocol,
-				  grub_efi_interface_type_t interface_type,
-				  void *interface);
+				  grub_efi_interface_type_t protocol_interface_type,
+				  void *protocol_interface);
 
   grub_efi_status_t
   (*reinstall_protocol_interface) (grub_efi_handle_t handle,
@@ -1026,12 +1126,12 @@ struct grub_efi_boot_services
   grub_efi_status_t
   (*uninstall_protocol_interface) (grub_efi_handle_t handle,
 				   grub_efi_guid_t *protocol,
-				   void *interface);
+				   void *protocol_interface);
 
   grub_efi_status_t
   (*handle_protocol) (grub_efi_handle_t handle,
 		      grub_efi_guid_t *protocol,
-		      void **interface);
+		      void **protocol_interface);
 
   void *reserved;
 
@@ -1107,7 +1207,7 @@ struct grub_efi_boot_services
   grub_efi_status_t
   (*open_protocol) (grub_efi_handle_t handle,
 		    grub_efi_guid_t *protocol,
-		    void **interface,
+		    void **protocol_interface,
 		    grub_efi_handle_t agent_handle,
 		    grub_efi_handle_t controller_handle,
 		    grub_efi_uint32_t attributes);
@@ -1126,7 +1226,7 @@ struct grub_efi_boot_services
 
   grub_efi_status_t
   (*protocols_per_handle) (grub_efi_handle_t handle,
-			   grub_efi_guid_t ***protocol_buffer,
+			   grub_efi_packed_guid_t ***protocol_buffer,
 			   grub_efi_uintn_t *protocol_buffer_count);
 
   grub_efi_status_t
@@ -1139,7 +1239,7 @@ struct grub_efi_boot_services
   grub_efi_status_t
   (*locate_protocol) (grub_efi_guid_t *protocol,
 		      void *registration,
-		      void **interface);
+		      void **protocol_interface);
 
   grub_efi_status_t
   (*install_multiple_protocol_interfaces) (grub_efi_handle_t *handle, ...);
@@ -1225,9 +1325,9 @@ typedef struct grub_efi_runtime_services grub_efi_runtime_services_t;
 
 struct grub_efi_configuration_table
 {
-  grub_efi_guid_t vendor_guid;
+  grub_efi_packed_guid_t vendor_guid;
   void *vendor_table;
-} __attribute__ ((packed));
+} GRUB_PACKED;
 typedef struct grub_efi_configuration_table grub_efi_configuration_table_t;
 
 #define GRUB_EFIEMU_SYSTEM_TABLE_SIGNATURE 0x5453595320494249LL
@@ -1244,7 +1344,8 @@ struct grub_efi_serial_io_interface
 				       grub_efi_parity_type_t parity,
 				       grub_uint8_t word_len,
 				       grub_efi_stop_bits_t stop_bits);
-  void (*set_control_bits) (void);
+  grub_efi_status_t (*set_control_bits) (struct grub_efi_serial_io_interface *this,
+					 grub_efi_uint32_t flags);
   void (*get_control_bits) (void);
   grub_efi_status_t (*write) (struct grub_efi_serial_io_interface *this,
 			      grub_efi_uintn_t *buf_size,
@@ -1268,6 +1369,43 @@ struct grub_efi_simple_input_interface
 };
 typedef struct grub_efi_simple_input_interface grub_efi_simple_input_interface_t;
 
+struct grub_efi_key_data {
+	grub_efi_input_key_t key;
+	grub_efi_key_state_t key_state;
+};
+typedef struct grub_efi_key_data grub_efi_key_data_t;
+
+typedef grub_efi_status_t (*grub_efi_key_notify_function_t) (
+	grub_efi_key_data_t *key_data
+	);
+
+struct grub_efi_simple_text_input_ex_interface
+{
+	grub_efi_status_t
+	(*reset) (struct grub_efi_simple_text_input_ex_interface *this,
+		  grub_efi_boolean_t extended_verification);
+
+	grub_efi_status_t
+	(*read_key_stroke) (struct grub_efi_simple_text_input_ex_interface *this,
+			    grub_efi_key_data_t *key_data);
+
+	grub_efi_event_t wait_for_key;
+
+	grub_efi_status_t
+	(*set_state) (struct grub_efi_simple_text_input_ex_interface *this,
+		      grub_efi_key_toggle_state_t *key_toggle_state);
+
+	grub_efi_status_t
+	(*register_key_notify) (struct grub_efi_simple_text_input_ex_interface *this,
+				grub_efi_key_data_t *key_data,
+				grub_efi_key_notify_function_t key_notification_function);
+
+	grub_efi_status_t
+	(*unregister_key_notify) (struct grub_efi_simple_text_input_ex_interface *this,
+				  void *notification_handle);
+};
+typedef struct grub_efi_simple_text_input_ex_interface grub_efi_simple_text_input_ex_interface_t;
+
 struct grub_efi_simple_text_output_interface
 {
   grub_efi_status_t
@@ -1475,17 +1613,31 @@ enum
     GRUB_EFI_NETWORK_INITIALIZED,
   };
 
+enum
+  {
+    GRUB_EFI_SIMPLE_NETWORK_RECEIVE_UNICAST		  = 0x01,
+    GRUB_EFI_SIMPLE_NETWORK_RECEIVE_MULTICAST		  = 0x02,
+    GRUB_EFI_SIMPLE_NETWORK_RECEIVE_BROADCAST		  = 0x04,
+    GRUB_EFI_SIMPLE_NETWORK_RECEIVE_PROMISCUOUS		  = 0x08,
+    GRUB_EFI_SIMPLE_NETWORK_RECEIVE_PROMISCUOUS_MULTICAST = 0x10,
+  };
+
 struct grub_efi_simple_network
 {
   grub_uint64_t revision;
   grub_efi_status_t (*start) (struct grub_efi_simple_network *this);
-  void (*stop) (void);
+  grub_efi_status_t (*stop) (struct grub_efi_simple_network *this);
   grub_efi_status_t (*initialize) (struct grub_efi_simple_network *this,
 				   grub_efi_uintn_t extra_rx,
 				   grub_efi_uintn_t extra_tx);
   void (*reset) (void);
-  void (*shutdown) (void);
-  void (*receive_filters) (void);
+  grub_efi_status_t (*shutdown) (struct grub_efi_simple_network *this);
+  grub_efi_status_t (*receive_filters) (struct grub_efi_simple_network *this,
+					grub_uint32_t enable,
+					grub_uint32_t disable,
+					grub_efi_boolean_t reset_mcast_filter,
+					grub_efi_uintn_t mcast_filter_count,
+					grub_efi_mac_address_t *mcast_filter);
   void (*station_address) (void);
   void (*statistics) (void);
   void (*mcastiptomac) (void);
@@ -1533,7 +1685,8 @@ struct grub_efi_block_io
 };
 typedef struct grub_efi_block_io grub_efi_block_io_t;
 
-#if (GRUB_TARGET_SIZEOF_VOID_P == 4) || defined (__ia64__)
+#if (GRUB_TARGET_SIZEOF_VOID_P == 4) || defined (__ia64__) \
+  || defined (__aarch64__) || defined (__MINGW64__) || defined (__CYGWIN__)
 
 #define efi_call_0(func)		func()
 #define efi_call_1(func, a)		func(a)
